#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "util.h"
#include "linked_list.h"
#include "mp3.h"
#include "id3_tag.h"
#include "id3_tag_frame.h"



/**
 * Global variables.
 */
	static bool   g_print_only 			  = false;
	static bool   g_rename				  = false;
	static bool   g_relocate			  = false;
	static size_t g_files_processed 	  = 0;
	static size_t g_directories_processed = 0;
	static size_t g_mean_tag_size 		  = 0;
	static const char** g_frames   		  = NULL;
	static int	  g_frame_count			  = 0;
	static char*  g_base_directory		  = NULL;
	extern const char PATH_SEPARATOR;



/**
 * Writes the specified mp3 file to the specified file descriptor.  Returns true if the file was successfully written, false otherwise.
 */
static bool write_file( const Mp3* mp3, int fd )
{

	// Validate args.
	//
		if( mp3 == NULL )
		{
			print_message( "Invalid argument: Mp3.", ERROR, false );
			return false;
		}


	// Reset to beginning of file.
	lseek( fd, SEEK_SET, 0 );

	size_t tag_size = v2_get_tag_storage_size( mp3->v2_tag );

	// Write the tag to file.
	//
		if( mp3->v2_tag != NULL )
		{
			// If after sanitization, the tag contains no frames, don't write the tag.
			if( list_size( mp3->v2_tag->frame_list ) < 1 )
				print_message( "After sanitization, tag contains no frames.  Not writing tag.  Please manually create a tag for this file.", WARNING, false );
			else
			{
				uint8_t* buffer = v2_get_tag_as_buffer( mp3->v2_tag );
				if( buffer == NULL || write( fd, buffer, tag_size ) != (ssize_t)tag_size )
				{
					print_message( "Error writing tag to file.", ERROR, true );
					lseek( fd, SEEK_SET, 0 );	// Reset and attempt to write data.
				}

				free( buffer );
			}
		}


	// Write mp3 data.
	//
		if( write( fd, mp3->data_offset, mp3->data_size ) != (ssize_t)mp3->data_size )
			print_message( "Error writing mp3 data to file.", ERROR, true );
		else
		{
			if( ftruncate( fd, tag_size + mp3->data_size ) == -1 )
				print_message( "Failed to truncate the file.", ERROR, true );
		}


	return true;
}



/**
 * Appends the specified integer value to the end of the specified mp3 file name.  The file argument should be a pointer created
 * by malloc.  The caller must pass the actual file pointer (not a copy), since file pointer will be reallocated to ensure enough
 * space is available to append the suffix.
 */
static void append_int_to_mp3_file( char** file, int i )
{

	// Validate args.
	//
		if( *file == NULL )
			return;

		int length = strlen( *file );
		if( length < 1 )
			return;


	// Allocate some space.
	//
		const int APPENDIX_LENGTH = 8;
		*file = realloc( *file, length + APPENDIX_LENGTH + 1 );


	// Find the point in the file name to start overwriting.
	//
		int index = length;
		if( length >= 8 )
		{
			if( (*file)[length-7] == '(' && (*file)[length-5] == ')' )
				index = length-7;
			else
				index = length-4;
		}


	// Append.
	//
		char appendix[5];
		sprintf( appendix, "(%i).mp3", i );
		strcpy( *file + index, appendix );
}



/**
 * Concatenates the two path components such that the result is 'path1 PATH_SEPARATOR path2'.  If the path cannot be created,
 * NULL is returned.  The returned string should be freed by the caller.
 */
static char* concat_path( const char* const path1, const char* const path2 )
{

	// Validate args.
	//
		if( path1 == NULL || path2 == NULL )
			return NULL;


	// Create path.
	//
		int size = strlen( path1 ) + strlen( path2 ) + 7;
		char* full_path = (char*)xcalloc( size, sizeof( char ) );
		strcpy( full_path, path1 );

		if( full_path[strlen( full_path )-1] != PATH_SEPARATOR )
			full_path[strlen( full_path )] = PATH_SEPARATOR;

		strcat( full_path, path2 );


	return full_path;
}



/**
 * Creates a fully qualified file path based on the base directory (provided by the user) and artist.
 */
static char* get_new_file_path( Mp3* mp3 )
{

	// Validate args.
	//
		if( mp3 == NULL )
			return NULL;


	// Get artist.
	//
		char* artist = get_artist( mp3 );

		if( artist == NULL )
		{
			print_message( "Could not determine artist.  File will not be moved.", WARNING, false );
			return NULL;
		}


	// If the directory does not exist then create it.
	//
		char* path = concat_path( g_base_directory, artist );
		free( artist );

		if( path == NULL )
		{
			print_message( "Could not create path.", WARNING, false );
			return NULL;
		}

		if( access( path, F_OK ) != 0 )
		{
			if( mkdir( path, S_IRWXU ) != 0 )
			{
				print_message( "Could not create directory.  File will not be moved.", ERROR, true );
				free( path );
				return NULL;
			}
		}


	return path;
}



/**
 * Processes a single file.
 */
static void process_file( const char* const file )
{

	// Validate args.
	//
		if( file == NULL )
			return;


	// Only process files with .mp3 extension.
	//
		if( (file[strlen( file )-3] != 'm' && file[strlen( file )-3] != 'M') ||
			(file[strlen( file )-2] != 'p' && file[strlen( file )-2] != 'P') ||
		     file[strlen( file )-1] != '3' )
		{
			printf( "\n\n\tSkipping non mp3 file: %s\n", file );
			return;
		}


	// Output some info to the user.
	//
		if( g_print_only )
			printf( "\n\n\nFile: %s", file );
		else
			printf( "\nProcessing file: %s", file );


	// Open the file.
	//
		int fd = open ( file, O_RDWR );
		if( fd < 0 )
		{
			print_message( "Could not open file.", ERROR, true );
			return;
		}


	// Create MP3_file instance.
	//
		Mp3* mp3 = get_mp3( file, fd, g_print_only );

		if( mp3 == NULL )
		{
			close( fd );
			return;
		}


	// Process the file.
	//
		if( g_print_only )
		{
			v1_print_tag( mp3->v1_tag );
			v2_print_tag( mp3->v2_tag );
		}
		else
		{
			// Rename the file.
			//
				if( g_rename )
					if( !rename_from_tag( mp3 ) )
						print_message( "Could not create file name from tag.  Not renaming file.", WARNING, false );


			// Relocate the file.
			//
				if( g_relocate )
				{
					char* temp_path = get_new_file_path( mp3 );

					if( temp_path != NULL )
					{
						free( mp3->file_path );
						mp3->file_path = temp_path;
					}
				}


			// Perform the actual rename.
			//
				// File might not have been renamed, and therefore won't have a new file name.
				char* full_path = NULL;
				if( mp3->new_file_name != NULL )
					full_path = concat_path( mp3->file_path, mp3->new_file_name );
				else
					full_path = concat_path( mp3->file_path, mp3->old_file_name );


				if( full_path == NULL )
					print_message( "Could not create fully qualified file path.  Not renaming or relocating file.", WARNING, false );
				else
				{
					// Before renaming, ensure we don't overwrite an existing file.
					//
						if( access( full_path, F_OK ) == 0 )	// File with the new name exists.
						{
							ino_t old_inode = get_inode_from_descriptor( fd );
							ino_t new_inode = get_inode_from_path( full_path );

							if( old_inode != new_inode ) // If the inode numbers don't match then the file names point to different files.
							{
								int i = 0;
								while( access( full_path, F_OK ) == 0 )
									append_int_to_mp3_file( &full_path, ++i );
							}
						}


					if( rename( file, full_path ) == -1 )
						print_message( "Rename failed.", ERROR, true );

					free( full_path );
				}


			// Sanitize the tag.
			//
				if( mp3->v2_tag != NULL )
					v2_sanitize_tag( mp3->v2_tag, g_frames, g_frame_count );


			// Write the file and set permission.
			//
				write_file( mp3, fd );
				fchmod( fd, S_IRUSR | S_IWUSR );
		}


	// Set metrics.
	//
		static long long tag_size_sum = 0;
		g_files_processed++;
		tag_size_sum += v2_get_tag_storage_size( mp3->v2_tag );
		g_mean_tag_size = tag_size_sum / g_files_processed;


	// Clean up.
	//
		free_mp3( mp3 );
		close( fd );
}



/**
 * Passed to scandir() for directory entry selection.
 */
static int directory_select( const struct dirent* dir_ent )
{

	if( (strlen( dir_ent->d_name ) <= 2) 	&&
		dir_ent->d_name[0] == '.' 			&&
		( dir_ent->d_name[1] == '.' || dir_ent->d_name[1] == 0 ) )
		return 0;

	if( dir_ent->d_type == DT_DIR || dir_ent->d_type == DT_REG  )
		return 1;

	return 0;
}



/**
 * Passed to scandir() for directory entry sorting.
 */
static int directory_sort( const struct dirent** dirent_1, const struct dirent** dirent_2 )
{

	// If both entries are directories or both are files, then sort alphabetically.
	//
		if( ((*dirent_1)->d_type == DT_DIR && (*dirent_2)->d_type == DT_DIR) ||
			((*dirent_1)->d_type == DT_REG && (*dirent_2)->d_type == DT_REG) )
			return alphasort( dirent_1, dirent_2 );


	// A directory entry takes precedence over a file entry.  Sort desc.
	//
		return (*dirent_1)->d_type == DT_DIR ? -1 : 1;
}



/**
 * Recursive method to process all directories in a tree.  A pre-order traversal is performed.
 */
static void process_directory( char* base_dir )
{

	// Validate args.
	//
		if( base_dir == NULL )
			return;


	// Add a trailing path separator if not present.
	//
		int length = strlen( base_dir );
		if( base_dir[length-1] != PATH_SEPARATOR )
		{
			base_dir[length]   = PATH_SEPARATOR;
			base_dir[length+1] = 0;
		}


	// Get a list of entries from the specified directory.
	//
		struct dirent** entries = NULL;
		int entry_count = scandir( base_dir, &entries, directory_select, directory_sort );

		for( int i = 0; i < entry_count; i++ )
		{
			// Need a new buffer for each call so we don't mess up current_path.
			length += strlen( entries[i]->d_name ) + 2;
			char path[length];
			strcpy( path, base_dir );
			strcat( path, entries[i]->d_name );


			if( entries[i]->d_type == DT_DIR )	// If the entry is a directory then recurse.
			{
				g_directories_processed++;
				process_directory( path );

				//If the directory is empty then delete it.
				//
					if( directory_is_empty( path ) )
						rmdir( path );
			}
			else if( entries[i]->d_type == DT_REG )	// If the entry is a regular file then process it.
				process_file( path );


			free( entries[i] );
		}


	// Clean up.
	//
		free( entries );
}



/**
 * Displays usage message.
 */
static void display_usage()
{
	printf( "\n\n Usage: bulkID3 [--print--only] [--rename] [--relocate] --frames=FRAME_ID_1[,FRAME_ID_2,...] directory \n\n" );

	const char* info = "\tWhere 'directory' is the directory containing the music files whose ID3 tags you wish to update.  "
					   "Directories are recursed by default."
					   "\n\tFor frame definitions see http://www.id3.org/id3v2.4.0-frames\n\n";

	printf( "%s", info );
	printf( "\n\n\t%-20s %s", "--print-only", "Do not modify tags, just display them." );
	printf( "\n\n\t%-20s %s", "--rename", "Use tag data to rename files." );
	printf( "\n\n\t%-20s %s", "--relocate", "Move each file to a sub-directory of 'directory', with the artist's name ( e.g. directory/artist/file.mp3 )." );
	printf( "\n\n\t%-20s %s", "--frames=frame-list", "List of valid ID3 frame types which should be retained in the tag ( e.g. --frames=TIT2,TPE1 ).  No spaces." );
	printf( "\n\t%-20s %s", "", "A wildcard (*) may be specified to indicate all frames should be retained ( e.g. --frames=* )." );
	printf( "\n\n\t%-20s %s", "--help", "Display this message.\n\n\n" );
}



/**
 * Entry point.
 */
int main( int argc, char** argv )
{

	// Process command line args.
	//
		if( argc < 2 )
		{
			display_usage();
			return EXIT_SUCCESS;
		}


		for( int i = 1; i < argc; i++ )
		{
			if( strncmp( argv[i], "--frames=", 9 ) == 0 )
			{
				// Tokenize the frame ID string.
				//
					// Make a copy since strtok() will modify the string.
					// Note: We won't free( frame_string ) as it will be necessary for the duration of the program's run.
					char* frame_string = strdup( argv[i] + 9 );

					g_frames = (const char**)split( frame_string, ',', &g_frame_count );

					if( g_frames == NULL )
					{
						display_usage();
						return EXIT_SUCCESS;
					}


				// Ensure the user entered valid frame IDs.
				//
					if( (*g_frames)[0] == '*' ) // Permit wildcard to specify all frames.
					{
						free( g_frames );
						g_frames = get_frame_list( &g_frame_count );
					}
					else
					{
						int invalid_frame_count = g_frame_count;
						const char** invalid_frames = validate_frames( (const char** const)g_frames, &invalid_frame_count);

						if( invalid_frames != NULL )
						{
							printf( "\nThe following frame IDs are invalid or unsupported: " );

							for( int j = 0; j < invalid_frame_count; j++ )
								printf( "%s ", invalid_frames[j] );

							printf( "\n\n" );
							return EXIT_SUCCESS;
						}
					}
			}
			else if( strcmp( argv[i], "--rename" ) == 0 )
			{
				g_rename = true;
			}
			else if( strcmp( argv[i], "--relocate" ) == 0 )
			{
				g_relocate = true;
			}
			else if( strcmp( argv[i], "--print-only" ) == 0 )
			{
				g_print_only = true;
			}
			else if( strcmp( argv[i], "--help" ) == 0 )
			{
				display_usage();
				return EXIT_SUCCESS;
			}
			else	// Assume anything else on the command line is a directory.
			{
				g_base_directory = argv[i];
			}
		}

		if( g_base_directory == NULL || (g_frames == NULL && !g_print_only) )	// opendir() will die if given a null pointer.
		{
			display_usage();
			return EXIT_SUCCESS;
		}



	// Ensure the specified directory is valid and start recursion.
	//
		DIR* dir = opendir( g_base_directory );
		if( dir == NULL )
		{
			printf( "\n Could not open directory: %s.\n", g_base_directory );
			return EXIT_SUCCESS;
		}

		closedir( dir );


	// Set buffering to line mode.
	//
		if( setvbuf( stdout, NULL, _IOLBF, 0 ) )
			printf( "\nCould not set stream buffering.  Output may be out of order\n." );


	// Start processing.
	//
		process_directory( g_base_directory );
		g_directories_processed++;	// Update here to account for the root directory.


	// Clean up
	//
		free( g_frames );


	// Output.
	//
		printf( "\n\n\nFiles processed: %li", g_files_processed );
		printf( "\nDirectories processed: %li", g_directories_processed );
		printf( "\nAverage tag size (bytes): %li\n\n", g_mean_tag_size );


	return EXIT_SUCCESS;
}
